package com.microsoft.notes

import android.content.Intent
import android.content.SharedPreferences
import android.os.Bundle
import android.preference.PreferenceManager
import android.support.design.widget.Snackbar
import android.support.v4.app.Fragment
import android.support.v7.widget.Toolbar
import android.util.Log
import android.view.Menu
import android.view.MenuItem
import android.view.MotionEvent
import android.view.View
import com.microsoft.notes.appstore.AuthMethods
import com.microsoft.notes.appstore.Navigation
import com.microsoft.notes.appstore.PreferenceKeys
import com.microsoft.notes.auth.AuthManager
import com.microsoft.notes.auth.AuthProvider
import com.microsoft.notes.components.ActivityResult
import com.microsoft.notes.components.RequestPermissionsResult
import com.microsoft.notes.components.UiMainActivityComponent
import com.microsoft.notes.components.ViewActivityWithControllerComponent
import com.microsoft.notes.editnote.EditNoteFragment
import com.microsoft.notes.extensions.addPhoto
import com.microsoft.notes.extensions.setupToolbar
import com.microsoft.notes.noteslist.NotesListFragment
import com.microsoft.notes.R
import com.microsoft.notes.noteslib.NotesLibrary
import com.microsoft.notes.search.SearchFragment
import com.microsoft.notes.store.Note
import com.microsoft.notes.transition.transitionCoordinator
import com.microsoft.notes.utils.logging.EventMarkers
import com.microsoft.notes.utils.logging.NotesOperation
import com.microsoft.notes.utils.logging.TelemetryKeys

open class ActivityStateManager(
        val controlledActivityComponent: ViewActivityWithControllerComponent<MainActivityControllerView>,
        val controlledActivityFragmentContainerID: Int,
        val controlledActivityAnimationOverLayViewID: Int) :  UiMainActivityComponent {

    companion object {
        const val TAG = "ActivityStateManager"
        const val EXTRA_MAIN_ACTIVITY_NOTE_ID = "main_activity_noteId"

        const val ACTION_VIEW_NOTES_LIST = "com.microsoft.notes.MainActivity.VIEW_NOTES_LIST"
        const val ACTION_ADD_TEXT_NOTE = "com.microsoft.notes.MainActivity.ADD_TEXT_NOTE"
        const val ACTION_ADD_PHOTO_NOTE = "com.microsoft.notes.MainActivity.ADD_PHOTO_NOTE"
        const val ACTION_EDIT_NOTE = "com.microsoft.notes.MainActivity.EDIT_NOTE"

        private const val STATE_CURRENT_FRAGMENT = "currentFragment"
        private const val STATE_CURRENT_NAV = "currentNav"
    }

    private var intentNoteId: String? = null
    private var authProvider: AuthProvider? = null

    protected val notesView = FragmentReference { NotesListFragment() }
    protected val editNote = FragmentReference { EditNoteFragment() }
    protected val search = FragmentReference { SearchFragment() }

    private val PREF_DEFAULT_VIEW: String by lazy {
        controlledActivityComponent.activity.getString(R.string
                .pref_default_view_key)
    }

    private val preferences: SharedPreferences
        get() = PreferenceManager.getDefaultSharedPreferences(controlledActivityComponent.activity)

    override var defaultView: Navigation
        get() {
            return Navigation.NOTES_LIST
        }
        set(value) {
            when (value) {
                Navigation.NOTES_LIST ->
                    preferences.edit().putString(PREF_DEFAULT_VIEW, value.name).apply()
                else -> IllegalArgumentException()
            }
        }

    private val currentFragment: Fragment?
        get() = when (controlledActivityComponent.cv.currentDisplayedNav) {
            Navigation.NOTES_LIST -> notesView.fragment
            Navigation.EDIT_NOTE -> editNote.fragment
            Navigation.SEARCH -> search.fragment
            else -> null
        }

    init {
        transitionCoordinator.animationOverlayViewID = controlledActivityAnimationOverLayViewID
    }

    fun onCreate(savedInstanceState: Bundle?) {
        controlledActivityComponent.registerForUpdates()
        parseIntent()
        bindFragments(savedInstanceState)
    }

    fun onResume() {
        controlledActivityComponent.startPolling()
    }

    fun onPause() {
        controlledActivityComponent.unregisterForUpdates()
        controlledActivityComponent.stopPolling()
    }

    fun onDestroy() {
        controlledActivityComponent.unregisterForUpdates()
    }

    fun onActivityResult(requestCode: Int, resultCode: Int, data: Intent?) {
        controlledActivityComponent.subActivityResultSubscribers.activityResults.onNext(
                ActivityResult(requestCode, resultCode, data))
        authProvider?.onResultAfterRequestToken(requestCode, resultCode, data)
    }

    fun onRequestPermissionsResult(requestCode: Int, permissions: Array<out String>,
                                   grantResults: IntArray) {
        controlledActivityComponent.subActivityResultSubscribers.permissionRequestResults.onNext(
                RequestPermissionsResult(requestCode, permissions, grantResults))
    }

    fun onSaveInstanceState(outState: Bundle) {
        currentFragment?.let {
            controlledActivityComponent.activity.supportFragmentManager.putFragment(outState,
                    STATE_CURRENT_FRAGMENT, it)
        }
        outState.putString(STATE_CURRENT_NAV, controlledActivityComponent.cv.currentDisplayedNav.name)
    }
    
    fun setupActionBar() {
        val notesToolbar = controlledActivityComponent.activity.findViewById<View>(R.id.notesToolbar)
        if(notesToolbar != null) {
            controlledActivityComponent.setupToolbar(notesToolbar as Toolbar)
        }
    }

    private fun parseIntent() {
        intentNoteId = controlledActivityComponent.activity.intent.getStringExtra(EXTRA_MAIN_ACTIVITY_NOTE_ID)
    }

    private fun bindFragments(savedInstanceState: Bundle?) {
        if (savedInstanceState == null) {
            val txn = controlledActivityComponent.activity.supportFragmentManager.beginTransaction()
            val defaultViewFragment: Fragment = notesView.fragment

            when (controlledActivityComponent.activity.intent.action) {
                ACTION_ADD_TEXT_NOTE -> {
                    controlledActivityComponent.cv.appLaunchedToEmptyNote()
                    txn.add(controlledActivityFragmentContainerID, editNote.fragment)
                }
                ACTION_ADD_PHOTO_NOTE -> {
                    controlledActivityComponent.cv.appLaunchedToDefaultView()
                    txn.add(controlledActivityFragmentContainerID, defaultViewFragment)
                    controlledActivityComponent.addPhoto {
                        controlledActivityComponent.cv
                                .createMediaNote(it)
                    }
                }
                ACTION_EDIT_NOTE -> {
                    intentNoteId?.let { controlledActivityComponent.cv.appLaunchedToEditNote(it) }
                            ?: controlledActivityComponent.cv.appLaunchedToEmptyNote()
                    txn.add(controlledActivityFragmentContainerID, editNote.fragment)
                }
                else -> {
                    controlledActivityComponent.cv.appLaunchedToDefaultView()
                    txn.add(controlledActivityFragmentContainerID, defaultViewFragment)
                }
            }

            txn.commit()
        } else {
            val currentNav = savedInstanceState.getString(STATE_CURRENT_NAV)
            val restoredFragment = controlledActivityComponent.activity.supportFragmentManager.getFragment(
                    savedInstanceState,
                    STATE_CURRENT_FRAGMENT)
            when (currentNav) {
                Navigation.NOTES_LIST.name -> {
                    notesView.restoreFragment(restoredFragment as NotesListFragment)
                }
                Navigation.EDIT_NOTE.name -> {
                    editNote.restoreFragment(restoredFragment as EditNoteFragment)
                }
                Navigation.SEARCH.name -> {
                    search.restoreFragment(restoredFragment as SearchFragment)
                }
                else -> {
                    Log.e(TAG,
                            "Unexpected Navigation in bindFragments: " + controlledActivityComponent.cv.currentDisplayedNav)
                }
            }
        }
    }

    fun onCreateOptionsMenu(menu: Menu): Boolean {
        controlledActivityComponent.activity.menuInflater.inflate(R.menu.main_menu, menu)

        return true
    }

    fun onPrepareOptionsMenu(menu: Menu) {
        menu.findItem(R.id.options_signin)?.isVisible = isSignedIn()
    }

    fun onOptionsItemSelected(item: MenuItem): Boolean {
        when (item.itemId) {
            R.id.actionSearch -> searchTriggered()
            R.id.actionDeleteNote -> deleteNoteTriggered()
            else -> return false
        }
        return true
    }

    fun onBackPressed(): Boolean {
        if (controlledActivityComponent.cv.previousStateNav == null) {
            return false
        }
        controlledActivityComponent.cv.navigatingBack()
        return true
    }

    fun onSwipe(isLeft: Boolean) {
        if(controlledActivityComponent.cv.currentDisplayedNav == Navigation.NOTES_LIST) {
                notesView.fragment.onSwipe(isLeft)
        }
    }

    fun dispatchTouchEvent(ev: MotionEvent?): Boolean? {
        if (controlledActivityComponent.cv.currentDisplayedNav == Navigation.EDIT_NOTE) {
            return editNote.fragment.gestureDetector?.onTouchEvent(ev)
        }
        return false
    }

    fun searchTriggered() {
        controlledActivityComponent.cv.recordTelemetry(EventMarkers.SearchTriggered,
                Pair(TelemetryKeys.CURRENT_NAVIGATION_STATE, controlledActivityComponent.cv.currentDisplayedNav.name))

        controlledActivityComponent.cv.navigatingToSearch()
    }

    override fun navigateTo(navigation: Navigation) =
            when (navigation) {
                Navigation.NOTES_LIST -> navigateToNotesList()
                Navigation.EDIT_NOTE -> navigateToEditNote(controlledActivityComponent.cv.currentNote)
                Navigation.SEARCH -> navigateToSearch()
                else -> Unit
            }

    private fun navigateToEditNote(note: Note?) {
        val txn = controlledActivityComponent.activity.supportFragmentManager.beginTransaction()
        editNote.recreate()

        when (controlledActivityComponent.cv.currentDisplayedNav) {
            Navigation.NOTES_LIST -> transitionCoordinator.notesListToEditNote(controlledActivityComponent.activity,
                    notesView.fragment,
                    editNote.fragment, txn, note)
            Navigation.SEARCH -> transitionCoordinator.searchToEditNote(controlledActivityComponent.activity,
                    search.fragment,
                    editNote.fragment,
                    txn, note)
            else -> transitionCoordinator.default(controlledActivityComponent.activity, currentFragment,
                    editNote.fragment,
                    txn)
        }

        txn.replace(controlledActivityFragmentContainerID, editNote.fragment).commit()
    }

    private fun navigateToNotesList() {
        val txn = controlledActivityComponent.activity.supportFragmentManager.beginTransaction()

        when (controlledActivityComponent.cv.currentDisplayedNav) {
            Navigation.EDIT_NOTE -> {
                transitionCoordinator.editNoteToNotesList(controlledActivityComponent.activity, editNote.fragment,
                        notesView.fragment,
                        txn)
            }
            Navigation.SEARCH -> txn.setCustomAnimations(R.animator.slide_from_left, R.animator.slide_to_right)
            else -> {
                transitionCoordinator.default(controlledActivityComponent.activity, currentFragment,
                        notesView.fragment,
                        txn)
            }
        }

        txn.replace(controlledActivityFragmentContainerID, notesView.fragment).commit()
    }

    private fun navigateToSearch() {
        val txn = controlledActivityComponent.activity.supportFragmentManager.beginTransaction()

        when (controlledActivityComponent.cv.currentDisplayedNav) {
            Navigation.NOTES_LIST -> txn.setCustomAnimations(R.animator.slide_from_right, R.animator.slide_to_left)
            Navigation.EDIT_NOTE -> transitionCoordinator.editNoteToSearch(controlledActivityComponent.activity,
                    editNote.fragment,
                    search.fragment, txn)
            else -> transitionCoordinator.default(controlledActivityComponent.activity, currentFragment,
                    search.fragment, txn)
        }

        txn.replace(controlledActivityFragmentContainerID, search.fragment).commit()
    }

    private fun deleteNoteTriggered() {
        if(controlledActivityComponent.cv.currentDisplayedNav == Navigation.EDIT_NOTE) {
            val currentNote: Note? = controlledActivityComponent.cv.currentNote
            if(currentNote != null) {
                controlledActivityComponent.cv.navigatingBack()
                archiveNote(currentNote)
            }
            else {
                Log.e(TAG, "Error while deleting, Note is null")
            }
        }
        else {
            Log.e(TAG, "Delete should only be called from Edit Note View")
        }
    }

    private fun archiveNote(note: Note)
    {
        controlledActivityComponent.cv.recordTelemetry(EventMarkers.NotesOperationTriggered,
                    Pair(TelemetryKeys.NOTES_OPERATION, NotesOperation.NoteArchived.name),
                    Pair(TelemetryKeys.TRIGGER_POINT, Navigation.EDIT_NOTE.name))
        NotesLibrary.getInstance().archiveNote(note.localId, note.remoteData?.id)
        Snackbar.make(controlledActivityComponent.activity.findViewById<View>(controlledActivityFragmentContainerID), R.string.label_archived, Snackbar.LENGTH_LONG)
                .setAction(R.string.snackbar_action_undo, {
                    unArchiveNote(note)
                })
                .setActionTextColor(controlledActivityComponent.activity.resources.getColor(R.color.link_text_darktheme))
                .show()
    }

    private fun unArchiveNote(note: Note)
    {
        controlledActivityComponent.cv.recordTelemetry(EventMarkers.NotesOperationTriggered,
                Pair(TelemetryKeys.NOTES_OPERATION, NotesOperation.NoteUnarchived.name),
                Pair(TelemetryKeys.TRIGGER_POINT, Navigation.EDIT_NOTE.name))
        NotesLibrary.getInstance().unArchiveNote(note.localId)
    }

    private fun isSignedIn(): Boolean {
        return preferences.getString(PreferenceKeys.AUTH_METHOD, "") != AuthMethods.AAD
    }

    override fun authenticate(method: String) {
        if (method == AuthMethods.AAD) {
            authProvider = AuthManager.getAuthProvider(controlledActivityComponent.activity, appStore = appStore)
            authProvider?.login(controlledActivityComponent.activity)
        } else {
            Log.d(TAG, "Unrecognized auth method")
        }
    }
}