package com.microsoft.notes.appstore

import com.microsoft.notes.appstore.action.Action
import com.microsoft.notes.appstore.action.AuthAction
import com.microsoft.notes.appstore.action.CompoundAction
import com.microsoft.notes.appstore.action.NavigationAction
import com.microsoft.notes.appstore.action.NoteUIAction
import com.microsoft.notes.appstore.action.ReadAction
import com.microsoft.notes.appstore.reducer.AuthReducer
import com.microsoft.notes.appstore.reducer.NavigationReducer
import com.microsoft.notes.appstore.reducer.NoteUIReducer
import com.microsoft.notes.appstore.reducer.ReadReducer
import com.microsoft.notes.utils.threading.ThreadExecutor
import java.util.concurrent.CopyOnWriteArrayList
import java.util.concurrent.LinkedBlockingQueue

abstract class AppStore {
    abstract val sideEffects: CopyOnWriteArrayList<SideEffect>
    abstract val stateHandlers: CopyOnWriteArrayList<StateHandler>

    var state = AppState()
        protected set

    abstract fun dispatch(action: Action)
    abstract fun dispatch(appState: AppState)
}

open class AppStoreImp(override val sideEffects: CopyOnWriteArrayList<SideEffect> = CopyOnWriteArrayList(),
                       override val stateHandlers: CopyOnWriteArrayList<StateHandler> = CopyOnWriteArrayList(),
                       val storeThread: ThreadExecutor? = null,
                       val logger: (String, String) -> Unit = { _, _ -> Unit }) : AppStore() {

    private var actions = LinkedBlockingQueue<Action>()

    @Synchronized
    override fun dispatch(action: Action) {
        actions.offer(action)
        if (storeThread != null) {
            storeThread.execute {
                handle(actions.poll())
            }
        } else {
            handle(actions.remove())
        }
    }

    private fun handle(action: Action) {
        val newState = reduce(action, state)
        dispatch(newState)
        sideEffects.dispatch(action)
    }

    override fun dispatch(appState: AppState) {
        this.state = appState
        stateHandlers.dispatch(appState)
    }

    private fun reduce(action: Action, currentAppState: AppState): AppState {
        logger(">>> ACTION", action.toString())
        val newState = when (action) {
            is CompoundAction -> {
                action.actions.fold(currentAppState) { state, nextAction ->
                    reduceSimpleAction(nextAction, state)
                }
            }
            else -> reduceSimpleAction(action, currentAppState)
        }
        logger(">>> STATE: ", newState.toString())
        return newState
    }

    private fun reduceSimpleAction(action: Action, currentAppState: AppState): AppState {
        logger(">>> SIMPLE ACTION: ", action.toString())
        return when (action) {
            is AuthAction -> AuthReducer.reduce(action, currentAppState)
            is NavigationAction -> NavigationReducer.reduce(action, currentAppState)
            is NoteUIAction -> NoteUIReducer.reduce(action, currentAppState)
            is ReadAction -> ReadReducer.reduce(action, currentAppState)
            else -> currentAppState
        }
    }
}

