package com.microsoft.notes.editnote

import android.content.Context
import android.net.Uri
import android.os.Bundle
import android.support.v4.view.GestureDetectorCompat
import android.support.v4.view.ViewCompat
import android.view.GestureDetector
import android.view.LayoutInflater
import android.view.Menu
import android.view.MenuInflater
import android.view.MotionEvent
import android.view.View
import android.view.ViewGroup
import android.view.inputmethod.InputMethodManager
import android.widget.EditText
import com.microsoft.notes.R
import com.microsoft.notes.ViewFragment
import com.microsoft.notes.appstore.getCurrentEditNote
import com.microsoft.notes.components.UiEditNotesFragmentComponent
import com.microsoft.notes.components.ViewActivityWithStateManagerComponent
import com.microsoft.notes.extensions.addPhoto
import com.microsoft.notes.extensions.setupToolbar
import com.microsoft.notes.extensions.toContextColor
import com.microsoft.notes.noteslist.TitleEditTextViewCallback
import com.microsoft.notes.richtext.editor.NotesEditTextCallback
import com.microsoft.notes.richtext.scheme.Document
import com.microsoft.notes.richtext.scheme.getUrlOrNull
import com.microsoft.notes.richtext.scheme.isEmpty
import com.microsoft.notes.richtext.scheme.mediaList
import com.microsoft.notes.store.Color
import com.microsoft.notes.store.Note
import com.microsoft.notes.utils.logging.CommandType
import com.microsoft.notes.utils.logging.EventMarkers
import com.microsoft.notes.utils.logging.TelemetryKeys
import kotlinx.android.synthetic.main.edit_note_content_layout.*
import kotlinx.android.synthetic.main.edit_note_layout.*
import java.lang.ref.WeakReference

class EditNoteFragment : ViewFragment<EditNoteControllerView>(),
        UiEditNotesFragmentComponent, AccessToViews, NotesEditTextCallback, TitleEditTextViewCallback {

    init {
        allowEnterTransitionOverlap = true
        allowReturnTransitionOverlap = true
    }

    override val accessToTitle: EditText
        get() = noteTitleEditText
    override val accessToBody: EditText
        get() = accessToBody

    var wasNoteLoaded: Boolean = false
    var gestureDetector: GestureDetectorCompat? = null

    override fun onCreateView(inflater: LayoutInflater, container: ViewGroup?, savedInstanceState: Bundle?): View {
        gestureDetector = GestureDetectorCompat(activity, EditNoteGestureDetector(WeakReference(this)))
        return inflater.inflate(R.layout.edit_note_layout, container, false)
    }

    override fun onActivityCreated(savedInstanceState: Bundle?) {
        super.onActivityCreated(savedInstanceState)
        setupTransitionNames()
        setupToolbar()
        bindClickEvents()
        setupEditNoteText()
        wasNoteLoaded = false
    }

    override fun setupControllerView() = EditNoteControllerView(uiFragmentComponent = this)

    private fun setupEditNoteText() {
        noteTitleEditText.setTitleEditTextViewCallback(this)
        noteBodyEditText.setNotesEditTextViewCallback(this)
    }

    fun onNavigateToTransitionCompleted() {
        if (cv?.currentNote?.isEmpty ?: true) {
            noteBodyEditText.requestFocus()
            val imm = activity.getSystemService(Context.INPUT_METHOD_SERVICE) as InputMethodManager
            imm.showSoftInput(noteBodyEditText, InputMethodManager.SHOW_IMPLICIT)
        }
    }

    private fun setupTransitionNames() {
        ViewCompat.setTransitionName(editNoteCardView, "card")
        ViewCompat.setTransitionName(scrollViewContainer, "linearLayout")
        ViewCompat.setTransitionName(noteTitleEditText, "title")
        ViewCompat.setTransitionName(noteBodyEditText, "body")
    }

    private fun setupToolbar() {
        parent().setupToolbar(notesToolbar)
        setHasOptionsMenu(true)
    }

    private fun bindClickEvents() {
        toolBarBackButton.setOnClickListener {
            activity.onBackPressed()
        }

        toggleBoldButton.setOnClickListener {
            recordCommand(CommandType.Bold)
            noteBodyEditText.toggleBold()
        }
        
        toggleItalicButton.setOnClickListener {
            recordCommand(CommandType.Italic)
            noteBodyEditText.toggleItalic()
        }
        
        toggleUnderlineButton.setOnClickListener {
            recordCommand(CommandType.Underline)
            noteBodyEditText.toggleUnderline()
        }
        
        toggleBulletsButton.setOnClickListener {
            recordCommand(CommandType.Bullets)
            noteBodyEditText.toggleBullets()
        }

        addPhotoButton.setOnClickListener {
            recordCommand(CommandType.AddPhoto)
            parent().addPhoto() { noteBodyEditText.addImage(it) }
        }
    }

    override fun onCreateOptionsMenu(menu: Menu?, inflater: MenuInflater?) {
        super.onCreateOptionsMenu(menu, inflater)

        menu?.findItem(R.id.actionSearch)?.isVisible = false
        menu?.findItem(R.id.actionNotebooksMode)?.isVisible = false
        menu?.findItem(R.id.options_notes_about)?.isVisible = false
        menu?.findItem(R.id.actionDeleteNote)?.isVisible = true
    }

    fun isEmpty() = noteTitleEditText.text.toString().isEmpty() &&
            noteBodyEditText.text.toString().isEmpty() &&
            (cv?.appState?.getCurrentEditNote()?.isDocumentEmpty ?: true)

    fun prepareSharedElements(markSharedElement: (View, String) -> Unit) {
        markSharedElement(editNoteCardView, "card")
        markSharedElement(scrollViewContainer, "linearLayout")
        markSharedElement(noteTitleEditText, "title")
        markSharedElement(noteBodyEditText, "body")
    }

    override fun parent(): ViewActivityWithStateManagerComponent = activity as ViewActivityWithStateManagerComponent

    override fun setNoteDetails(note: Note) =
            with(note) {
                noteTitleEditText.setTitleWhenDifferent(title)
                noteBodyEditText.setDocumentWithDelayedImages(
                        document = document)
                setSideNoteColor(color)
                changeTitleVisibilityWhenNoteLoadedFirstTime(this)
            }

    private fun changeTitleVisibilityWhenNoteLoadedFirstTime(note: Note) {
        //FIXME Don't really like this flag, I will explore more options.
        if (!wasNoteLoaded) {
            applyWithoutTransition {
                noteTitleEditText.setTitleVisibilityWhenThereIsOnlyBodyText(note)
            }
            wasNoteLoaded = true
        }
    }

    fun couldHideTitle(): Boolean {
        val noteTitleVisibility = noteTitleEditText?.visibility ?: false
        val isNoteTitleEmpty = noteTitleEditText?.text?.isEmpty() ?: false

        return noteTitleVisibility == View.VISIBLE && isNoteTitleEmpty
    }

    fun hideTitle() {
        if (couldHideTitle()) {
            noteTitleEditText.visibility = View.GONE
        }
    }

    override fun updateDocument(document: Document) {
        cv?.updateNoteWithDocument(document = document)
        if (!document.isEmpty())
            hideTitle()
    }

    override fun updateTitle(title: String) {
        cv?.updateNoteWithTitle(titleText = title)
    }

    override fun moveFocusToBody() {
        noteBodyEditText.requestFocus()
    }

    override fun openMediaInFullScreen(mediaLocalId: String) {
        recordCommand(CommandType.OpenMedia)
        val mediaUri = getMediaUriFromLocalId(mediaLocalId)
        if(mediaUri != null)
        {
            parent().viewPhotoInFullScreen(mediaUri.path)
        }
    }

    private fun recordCommand(commandType: CommandType, vararg keyValuePairs: Pair<String, String>) {
        cv?.recordTelemetry(EventMarkers.CommandTriggered, Pair(TelemetryKeys.COMMAND_TYPE, commandType.name),
                *keyValuePairs)
    }

    private fun setSideNoteColor(color: Color) {
        editNoteCardView.setCardBackgroundColor(color.toContextColor(context))
    }

    private fun applyWithoutTransition(function: () -> Unit) {
        val layoutTransition = editNoteLayout.layoutTransition
        editNoteLayout.layoutTransition = null
        function()
        editNoteLayout.layoutTransition = layoutTransition
    }
    
    private fun getMediaUriFromLocalId(mediaLocalId: String) : Uri?
    {
        val mediaList = cv?.currentNote?.document?.mediaList()?: emptyList()
        val imageIndex = Math.max(mediaList.indexOfFirst { it.localId == mediaLocalId }, 0)
        if(imageIndex < mediaList.size && mediaList[imageIndex].getUrlOrNull() != null)
            return Uri.parse(mediaList[imageIndex].getUrlOrNull())

        return null
    }
}

private interface AccessToViews {
    val accessToTitle: EditText
    val accessToBody: EditText
}

private class EditNoteGestureDetector(val accessToViews: WeakReference<AccessToViews>) :
        GestureDetector.SimpleOnGestureListener() {

    override fun onDown(e: MotionEvent): Boolean = true

    override fun onFling(e1: MotionEvent, e2: MotionEvent, velocityX: Float, velocityY: Float): Boolean {
        with(accessToViews) {
            if (e1.rawY < e2.rawY) {
                get()?.let { it.accessToTitle.visibility = View.VISIBLE }
            }
            return true
        }
    }
}
