package com.microsoft.notes.extensions

import android.app.Activity
import android.content.pm.ActivityInfo
import android.net.Uri
import android.support.v7.widget.Toolbar
import android.widget.Toast
import android.widget.Toast.LENGTH_LONG
import com.microsoft.notes.R
import com.microsoft.notes.components.ActivityResult
import com.microsoft.notes.components.NotesViewNavigation.PICTURES_DIRECTORY
import com.microsoft.notes.components.ViewActivityComponent
import com.microsoft.notes.photos.*
import com.microsoft.office.onenote.utils.ONMGalleryUtils
import com.microsoft.office.onenote.utils.ONMStringUtils
import rx.Observable
import java.io.File

fun ViewActivityComponent.createImageFileAsync(): Observable<File> = createImageFileAsync(activity)

fun ViewActivityComponent.addPhoto(photoHandler: (String) -> Unit) {
    PhotoPickerDialog().showPictureLauncher(activity) {
        val activityInfo = it.intentResolveInfo.activityInfo
        when (it.type) {
            PhotoPickerType.CAMERA -> takePhoto(activityInfo, photoHandler)
            PhotoPickerType.GALLERY -> pickPhoto(activityInfo, photoHandler)
        }
    }
}

fun ViewActivityComponent.quickCapturePhoto(photoHandler: (String) -> Unit) {
    val cameraActivityInfo : ActivityInfo? = getDefaultCameraActivityInfo(activity)
    if(cameraActivityInfo != null) {
        takePhoto(cameraActivityInfo, photoHandler, CameraSlideInDirection.LEFT /* slide in from left */)
    }
}

fun ViewActivityComponent.takePhoto(activityInfo: ActivityInfo, photoHandler: (String) -> Unit) {
    takePhoto(activityInfo, photoHandler, CameraSlideInDirection.NONE)
}

fun ViewActivityComponent.takePhoto(activityInfo: ActivityInfo, photoHandler: (String) -> Unit, slideDirection: CameraSlideInDirection) {

    fun showStoragePermissionDeniedMessage() {
        Toast.makeText(activity, R.string.label_camera_permission_denied, Toast.LENGTH_SHORT).show()
    }

    fun fileCreatedCallback(file: File) {

        fun activityResultCallback(result: ActivityResult) {
            if (result.resultCode == Activity.RESULT_OK) {
                val uri = Uri.fromFile(file)
                photoHandler(uri.toString())
                activity.addMediaToGallery(uri)
            } else {
                file.deleteFileAsync()
            }
        }

        takePhoto(this, file, activityInfo.applicationInfo.packageName,
                activityInfo.name, slideDirection, permissionDeniedCallback = ::showStoragePermissionDeniedMessage, requestCompletedCallback = ::activityResultCallback)
    }

    createImageFileAsync().subscribe(::fileCreatedCallback)
}

fun ViewActivityComponent.pickPhoto(activityInfo: ActivityInfo, photoHandler: (String) -> Unit) {

    fun showStoragePermissionDeniedSnackbar() {
        Toast.makeText(activity, R.string.label_external_storage_permission_denied, LENGTH_LONG).show()
    }

    fun activityResultCallback(result: ActivityResult) {
        if (result.resultCode == Activity.RESULT_OK) {
            val path = copyFileToLocal(result.intent?.data!!)
            if(path != null)
                photoHandler(Uri.fromFile(File(path)).toString())
        }
    }

    pickPhoto(this, activityInfo.applicationInfo.packageName, activityInfo.name,
            permissionDeniedCallback = ::showStoragePermissionDeniedSnackbar,
            requestCompletedCallback = ::activityResultCallback)
}

fun ViewActivityComponent.copyFileToLocal(uri : Uri) : String? {
    val localDir = activity.filesDir.path + File.separator + PICTURES_DIRECTORY
    return ONMGalleryUtils.getLocalFilePathAsActivityResult(activity, uri, localDir, ONMStringUtils.EMPTY_STRING /* intune identity*/)
}

fun ViewActivityComponent.setupToolbar(toolbar: Toolbar) {
    if (showNotesActionBar) {
        activity.setSupportActionBar(toolbar)
    }
}

fun ViewActivityComponent.setToolbarTitle(title: Int) {
    activity.supportActionBar?.setTitle(title)
}