package com.microsoft.notes.extensions

import android.graphics.Bitmap
import android.graphics.Point
import android.graphics.PointF
import android.graphics.Rect
import android.support.v4.graphics.drawable.RoundedBitmapDrawableFactory
import android.transition.Transition
import android.transition.TransitionManager
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.view.ViewTreeObserver
import android.widget.EditText
import android.widget.FrameLayout
import android.widget.ImageView
import android.widget.TextView
import com.bumptech.glide.Glide
import com.bumptech.glide.request.target.BitmapImageViewTarget
import com.microsoft.notes.R
import java.util.ConcurrentModificationException
import kotlin.reflect.KClass

fun EditText.moveCursorToTheEnd() = this.setSelection(this.text.length)

fun EditText.isTextDifferent(newText: String): Boolean = this.text.toString() != newText

fun EditText.setTextWhenDifferent(text: String?) =
        text?.let {
            if (this.isTextDifferent(it)) {
                this.setText(it)
                this.moveCursorToTheEnd()
            }
        }

/**
 * Set the given {@link content} into the {@link textView} when is not null or not empty, and set this view to
 * {@link VISIBLE}, otherwise the view visibility will set to {@link GONE}
 */
fun TextView.setContentAndVisibility(content: String?) {
    if (content != null && content.isNotEmpty()) {
        this.text = content
        this.visibility = View.VISIBLE
    } else {
        this.visibility = View.GONE
    }
}

/**
 * Set the given {@link content} after truncation into the {@link textView} when is not null or not empty, and set
 * this view to
 * {@link VISIBLE}, otherwise the view visibility will set to {@link GONE}
 */
fun TextView.setPreviewContentAndVisibility(content: String?, maxLength: Int) {
    if (content != null && content.isNotEmpty()){
        if(content.length > maxLength) {
            this.text = content.substring(0, maxLength)
        } else {
            this.text = content
        }
        this.visibility = View.VISIBLE
    } else {
        this.visibility = View.GONE
    }

}

fun ImageView.loadRoundedImageFromUri(uri: String?, radius: Float) {
    if (!uri.isNullOrEmpty()) {
        Glide.with(context).load(uri).asBitmap().centerCrop().into(object : BitmapImageViewTarget(this) {
            override fun setResource(resource: Bitmap) {
                val circularBitmapDrawable = RoundedBitmapDrawableFactory.create(context.resources, resource)
                circularBitmapDrawable.cornerRadius = radius
                setImageDrawable(circularBitmapDrawable)
            }
        })
    }
}

fun ImageView.loadImageFromUri(uri: String?) {
    if (!uri.isNullOrEmpty()) {
        Glide.with(context).load(uri).into(this)
    }
}

fun Iterable<View>.setVisibility(visibility: Int) = forEach { it.visibility = visibility }

/** TODO We have to give attribution to Anko library for these three functions in the case we ship UI and we use the
functions, right now our UI is using them:
Source code: https://github.com/Kotlin/anko/blob/4584577b6bf3b4f8d8d1733f53408537e2aac759/anko/library/static/commons/src/viewChildrenSequences.kt
License: Apache License v2, Copyright 2016 JetBrains s.r.o.
 */
fun View.childrenRecursiveSequence(): Sequence<View> = ViewChildrenRecursiveSequence(this)

fun View.childrenSequence(): Sequence<View> = ViewChildrenSequence(this)

private class ViewChildrenSequence(private val view: View) : Sequence<View> {
    override fun iterator(): Iterator<View> {
        if (view !is ViewGroup) return emptyList<View>().iterator()
        return ViewIterator(view)
    }

    private class ViewIterator(private val view: ViewGroup) : Iterator<View> {
        private var index = 0
        private val count = view.childCount

        override fun next(): View {
            if (!hasNext()) throw NoSuchElementException()
            return view.getChildAt(index++)
        }

        override fun hasNext(): Boolean {
            checkCount()
            return index < count
        }

        private fun checkCount() {
            if (count != view.childCount) throw ConcurrentModificationException()
        }
    }
}

private class ViewChildrenRecursiveSequence(private val view: View) : Sequence<View> {
    override fun iterator(): Iterator<View> {
        if (view !is ViewGroup) return emptyList<View>().iterator()
        return RecursiveViewIterator(view)
    }

    private class RecursiveViewIterator(view: View) : Iterator<View> {
        private val sequences = arrayListOf(view.childrenSequence())
        private var current = sequences.removeLast().iterator()

        override fun next(): View {
            if (!hasNext()) throw NoSuchElementException()
            val view = current.next()
            if (view is ViewGroup && view.childCount > 0) {
                sequences.add(view.childrenSequence())
            }
            return view
        }

        override fun hasNext(): Boolean {
            if (!current.hasNext() && sequences.isNotEmpty()) {
                current = sequences.removeLast().iterator()
            }
            return current.hasNext()
        }

        @Suppress("NOTHING_TO_INLINE")
        private inline fun <T : Any> MutableList<T>.removeLast(): T {
            if (isEmpty()) throw NoSuchElementException()
            return removeAt(size - 1)
        }
    }
}
//TODO End of Anko functions --- //

fun View.findViewsById(id: Int) = childrenRecursiveSequence().filter { this.id == id }

fun View.findViewsByType(clazz: KClass<*>) = childrenRecursiveSequence().filter { clazz.isInstance(it) }

fun View.nextLayoutListener(listener: () -> Unit) {
    viewTreeObserver.addOnGlobalLayoutListener(
            object : ViewTreeObserver.OnGlobalLayoutListener {
                override fun onGlobalLayout() {
                    listener()
                    viewTreeObserver.removeOnGlobalLayoutListener(this)
                }
            })
}

fun View.getLocationOnScreen(): IntArray {
    val position = IntArray(2)
    getLocationOnScreen(position)
    return position
}

fun View.getLocationOnScreenAsPoint(): Point {
    val position = getLocationOnScreen()
    return Point(position[0], position[1])
}

fun View.getBoundsOnScreen(): Rect {
    val position = getLocationOnScreen()
    return Rect(position[0], position[1], position[0] + width, position[1] + height)
}

fun View.setLayoutParamsFrom(other: View, parentPosition: IntArray) {
    val position = other.getLocationOnScreen()

    val layoutParams = FrameLayout.LayoutParams(layoutParams)
    layoutParams.leftMargin = position[0] - parentPosition[0]
    layoutParams.topMargin = position[1] - parentPosition[1]
    layoutParams.height = other.height
    layoutParams.width = other.width
    this.layoutParams = layoutParams
}

fun View.setLayoutParamsFrom(other: View, parentPosition: Point) {
    val position = other.getLocationOnScreen()

    val layoutParams = FrameLayout.LayoutParams(layoutParams)
    layoutParams.leftMargin = position[0] - parentPosition.x
    layoutParams.topMargin = position[1] - parentPosition.y
    layoutParams.height = other.height
    layoutParams.width = other.width
    this.layoutParams = layoutParams
}

fun ViewGroup.replaceChild(layoutId: Int, transition: Transition?): View {
    if (transition != null) {
        TransitionManager.beginDelayedTransition(this, transition)
    }
    removeAllViews()
    val view = LayoutInflater.from(context).inflate(layoutId, this, false)
    addView(view)
    return view
}

operator fun Rect.minus(point: IntArray) =
        Rect(left - point[0], top - point[1], right - point[0], bottom - point[1])

operator fun Rect.minus(point: Point) =
        Rect(left - point.x, top - point.y, right - point.x, bottom - point.y)

fun Rect.center() = Point(centerX(), centerY())

operator fun Point.plus(other: Point) = Point(x + other.x, y + other.y)

operator fun Point.minus(other: Point) = Point(x - other.x, y - other.y)

operator fun PointF.minus(other: PointF) = PointF(x - other.x, y - other.y)
