package com.microsoft.notes.noteslist

import android.content.Context
import android.os.Build
import android.support.annotation.RequiresApi
import android.support.design.widget.Snackbar
import android.support.v4.app.FragmentTransaction
import android.support.v7.widget.LinearLayoutManager
import android.util.AttributeSet
import android.view.LayoutInflater
import android.widget.FrameLayout
import com.microsoft.notes.R
import com.microsoft.notes.components.ViewActivityComponent
import com.microsoft.notes.appstore.ScrollTo
import com.microsoft.notes.extensions.getStableId
import com.microsoft.notes.noteslist.recyclerview.NoteItemComponent
import com.microsoft.notes.noteslist.recyclerview.NoteViewHolder
import com.microsoft.notes.noteslist.recyclerview.NotesListAdapter
import com.microsoft.notes.store.Color
import com.microsoft.notes.store.Note
import kotlinx.android.synthetic.main.notes_list_component_layout.view.*

class NotesListComponent(context: Context, attributeSet: AttributeSet?) : FrameLayout(context, attributeSet) {
    companion object {
        const val EXPAND_FROM_TOP = -1
    }

    constructor(context: Context) : this(context, null)

    var notesCollection: List<Note> = emptyList()
        private set

    private val notesAdapter: NotesListAdapter
    var callbacks: Callbacks? = null

    private val isUnderTest: Boolean
        get() = (context as? ViewActivityComponent)?.isUnderTest() ?: false

    val recyclerViewID: Int
        get() = R.id.notesRecyclerView

    // This value is used by the NotesListAdapter to set default names on the item views when we are returning
    // to the notes list from the details view, so that way the appropriate views are either part of the shared
    // element default. We lazily compute its value whenever we are starting this default, and otherwise its
    // value is meaningless.
    private var expandedPositionForReturnTransition: Lazy<Int> = lazy { findExpandedPositionFromNotesList() }

    init {
        LayoutInflater.from(context).inflate(R.layout.notes_list_component_layout, this)

        notesAdapter = NotesListAdapter(
                notes = emptyList<Note>(),
                getExpandedPosition = { expandedPositionForReturnTransition.value },
                itemClick = { callbacks?.onNoteClicked(it) },
                setColor = { note, color -> callbacks?.onNoteColorChanged(note, color) },
                updateNotesPositions = { notes -> callbacks?.onNotesReordered(notes) })

        notesRecyclerView.itemAnimator.changeDuration = 0
        val linearLayoutManager = LinearLayoutManager(context)
        notesRecyclerView.layoutManager = linearLayoutManager
        notesRecyclerView.adapter = notesAdapter

        setupSwipeToRefresh()
    }

    private fun setupSwipeToRefresh() {
        //TODO Implement here the trigger to sync new data from api
        notesSwipeToRefresh.setOnRefreshListener { notesSwipeToRefresh.isRefreshing = false }
    }

    fun showProgressBar() {
        if (!isUnderTest) notesSwipeToRefresh.isRefreshing = true
    }

    fun hideProgressBar() {
        if (!isUnderTest) notesSwipeToRefresh.isRefreshing = false
    }

    fun findNoteView(note: Note): NoteViewHolder? =
            notesRecyclerView.findViewHolderForItemId(note.getStableId()) as NoteViewHolder?

    fun prepareForNewNoteTransition() {
        setTransitionSlideDirections(EXPAND_FROM_TOP)
    }

    @RequiresApi(Build.VERSION_CODES.LOLLIPOP)
    fun prepareForEditNoteTransition(note: Note, txn: FragmentTransaction): NoteItemComponent? {
        val vh = notesRecyclerView.findViewHolderForItemId(note.getStableId()) as NoteViewHolder? ?: return null

        setTransitionSlideDirections(vh.layoutPosition)
        vh.noteView.prepareSharedElements { view, name ->
            view.transitionName = name
            txn.addSharedElement(view, name)
        }
        return vh.noteView
    }

    fun prepareForReturnTransition() {
        expandedPositionForReturnTransition = lazy { findExpandedPositionFromNotesList() }
    }

    private fun setTransitionSlideDirections(position: Int) {
        for (i: Int in 0 .. notesAdapter.itemCount) {
            val viewHolder = notesRecyclerView.findViewHolderForLayoutPosition(i) as NoteViewHolder?
            viewHolder?.noteView?.let {
                when {
                    i < position -> it.setRootTransitionName("up")
                    i > position -> it.setRootTransitionName("down")
                }
            }
        }
    }

    private fun findExpandedPositionFromNotesList(): Int {
        val currentNote = callbacks?.noteForReturnTransition ?: return EXPAND_FROM_TOP
        return if (currentNote.isEmpty) {
            EXPAND_FROM_TOP
        } else {
            val currentNoteId = currentNote.localId
            notesCollection.indexOfFirst { it.localId == currentNoteId }
        }
    }

    fun updateNotesCollection(notesCollection: List<Note>, scrollTo: ScrollTo) {
        this.notesCollection = notesCollection
        notesAdapter.updateNotesCollection(notesCollection)
        when (scrollTo) {
            is ScrollTo.Bottom -> scrollToPosition(position = notesCollection.size - 1, offset = 20)
            is ScrollTo.Top -> scrollToPosition(position = 0)
            is ScrollTo.Custom -> scrollToPosition(position = scrollTo.position)
            is ScrollTo.NoScroll -> Unit
        }
        hideProgressBar()
    }

    private fun scrollToPosition(position: Int, offset: Int = Int.MIN_VALUE) {
        (notesRecyclerView.layoutManager as LinearLayoutManager).scrollToPositionWithOffset(position, offset)
    }

    open class Callbacks {
        open val noteForReturnTransition: Note? get() = null
        open fun onNoteColorChanged(note: Note, color: Color) {}
        open fun onNotesReordered(notesList: List<Note>) {}
        open fun onNoteClicked(note: Note) {}
    }
}