package com.microsoft.notes.noteslist

import android.os.Build
import android.os.Bundle
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.view.ViewTreeObserver
import com.microsoft.notes.R
import com.microsoft.notes.ViewFragment
import com.microsoft.notes.appstore.Navigation
import com.microsoft.notes.appstore.ScrollTo
import com.microsoft.notes.appstore.getCurrentEditNote
import com.microsoft.notes.components.UiNotesFragmentComponent
import com.microsoft.notes.components.ViewActivityWithStateManagerComponent
import com.microsoft.notes.extensions.quickCapturePhoto
import com.microsoft.notes.extensions.setupToolbar
import com.microsoft.notes.noteslist.recyclerview.NoteViewHolder
import com.microsoft.notes.store.Color
import com.microsoft.notes.store.Note
import com.microsoft.notes.utils.logging.EventMarkers
import com.microsoft.notes.utils.logging.NoteType
import com.microsoft.notes.utils.logging.NotesOperation
import com.microsoft.notes.utils.logging.TelemetryKeys
import kotlinx.android.synthetic.main.notes_layout.*
import kotlinx.android.synthetic.main.notes_list_component_layout.view.*

class NotesListFragment : ViewFragment<NotesListControllerView>(), UiNotesFragmentComponent {

    init {
        allowEnterTransitionOverlap = true
        allowReturnTransitionOverlap = true
    }

    override fun onCreateView(inflater: LayoutInflater, container: ViewGroup?, savedInstanceState: Bundle?): View {
        return inflater.inflate(R.layout.notes_layout, container, false)
    }

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)
        notesList.notesRecyclerView.viewTreeObserver.addOnGlobalLayoutListener(object : ViewTreeObserver.OnGlobalLayoutListener {
            override fun onGlobalLayout() {
                notesList.notesRecyclerView.viewTreeObserver.removeOnGlobalLayoutListener(this)
                parent().onFragmentLoadComplete(view.id)
            }
        })
    }

    override fun onActivityCreated(savedInstanceState: Bundle?) {
        super.onActivityCreated(savedInstanceState)

        setupToolbar()
        setupNotesList()
        setupAddNote()
    }

    override fun setupControllerView() = NotesListControllerView(uiFragmentComponent = this)

    private fun setupToolbar() {
        parent().setupToolbar(notesToolbar)

        toolBarCameraButton.setOnClickListener {
            parent().quickCapturePhoto { cv?.createAndNavigateToMediaNote(it) }
            //TODO Add telemetry marker
        }
    }

    private fun setupNotesList() {
        notesList.callbacks = object : NotesListComponent.Callbacks() {

            override val noteForReturnTransition: Note?
                get() = cv?.appState?.getCurrentEditNote()

            override fun onNoteColorChanged(note: Note, color: Color) {
                recordNotesOperation(NotesOperation.NoteColorChanged)
                cv?.setNoteColor(note, color)                
            }

            override fun onNotesReordered(notesList: List<Note>) {
                recordNotesOperation(NotesOperation.NotesReordered)
                cv?.reorderNotes(notesList)                
            }

            override fun onNoteClicked(note: Note) {
                recordNotesOperation(NotesOperation.NoteClicked)
                editNoteWithContent(note)
            }
        }
    }

    private fun recordNotesOperation(notesOperation: NotesOperation) {
        cv?.recordTelemetry(EventMarkers.NotesOperationTriggered,
                Pair(TelemetryKeys.NOTES_OPERATION, notesOperation.name),
                Pair(TelemetryKeys.TRIGGER_POINT, Navigation.NOTES_LIST.name))
    }

    private fun setupAddNote() {
        newNoteFab.setOnClickListener { 
            cv?.recordTelemetry(EventMarkers.AddNoteTriggered, Pair(TelemetryKeys.NOTE_TYPE, NoteType.Default.name))
            editNewNote()
        }
        setupAddNoteAccessibility()
    }

    private fun setupAddNoteAccessibility() {
        if(Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP_MR1)
        {
            newNoteFab.accessibilityTraversalAfter = notesToolbar.id
            newNoteFab.accessibilityTraversalBefore = notesList.recyclerViewID
        }
    }

    override fun parent(): ViewActivityWithStateManagerComponent = activity as ViewActivityWithStateManagerComponent

    fun findCurrentNoteView(): NoteViewHolder? =
            cv?.appState?.getCurrentEditNote()?.let { notesList.findNoteView(it) }

    fun editNewNote(paragraphText: String? = null) {
        cv?.createAndNavigateToNote(paragraphText)
    }

    private fun editNoteWithContent(note: Note) {
        cv?.navigatingToNote(note)
    }

    override fun updateNotesCollection(notesCollection: List<Note>, scrollTo: ScrollTo) {
        notesList.updateNotesCollection(notesCollection, scrollTo)
    }

    fun onSwipe(isLeft : Boolean) {
        if(isLeft) {
            parent().quickCapturePhoto { cv?.createAndNavigateToMediaNote(it) }
        }
        else {
            cv?.navigatingToSearch()
        }
    }
}