package com.microsoft.notes.noteslist.recyclerview

import android.content.Context
import android.support.v4.content.ContextCompat
import android.support.v4.view.ViewCompat
import android.support.v7.widget.CardView
import android.support.v7.widget.RecyclerView
import android.support.v7.widget.ViewUtils
import android.util.AttributeSet
import android.util.TypedValue
import android.view.View
import android.widget.ImageButton
import android.widget.ImageView
import android.widget.LinearLayout
import android.widget.TextView
import com.microsoft.notes.R
import com.microsoft.notes.extensions.formatString
import com.microsoft.notes.extensions.loadRoundedImageFromUri
import com.microsoft.notes.extensions.setContentAndVisibility
import com.microsoft.notes.extensions.setPreviewContentAndVisibility
import com.microsoft.notes.extensions.toContextColor
import com.microsoft.notes.richtext.scheme.asString
import com.microsoft.notes.richtext.scheme.getUrlOrNull
import com.microsoft.notes.richtext.scheme.mediaList
import com.microsoft.notes.store.Color
import com.microsoft.notes.store.Note

class NoteViewHolder(noteView: NoteItemComponent) : RecyclerView.ViewHolder(noteView), ItemTouchHelperViewHolder {
    val noteView
        get() = itemView as NoteItemComponent

    val sourceNote
        get() = noteView.sourceNote

    override fun onItemSelected() {
        with(noteView) {
            val typedValue = TypedValue()
            context.theme.resolveAttribute(R.attr.themeColorContentBackgroundPressed, typedValue, true)
            contentLayout.setBackgroundColor(typedValue.data)
        }
    }

    override fun onItemClear() {
        with(noteView) {
            val typedValue = TypedValue()
            context.theme.resolveAttribute(R.attr.themeDrawableItemBackground, typedValue, true)
            contentLayout.background = ContextCompat.getDrawable(context, typedValue.resourceId)
            callbacks?.onNotesReordered()
        }
    }
}

open class NoteItemComponent(context: Context, attrs: AttributeSet) : CardView(context, attrs) {

    companion object {
        val INITIAL_DX_VALUE: Float = 0f
        //We dont want to set very large content on the textview as we are limiting it to 4 lines
        val MAX_CHARACTER_COUNT_PREVIEW = 500
    }

    val contentLayout by lazy { findViewById<View>(R.id.noteContentLayout) }
    lateinit var sourceNote: Note
    val colorPickerMap = mutableMapOf<ImageButton, Color>()

    var callbacks: Callbacks? = null

    private val isRTL: Boolean
        get() = ViewUtils.isLayoutRtl(this)

    open fun bindNote(note: Note, shouldShowColorPicker: Boolean) {
        with(note) {
            sourceNote = this
            bindClickHandlers()
            setContentLayoutBackgroundColor(sourceNote.color)
        }
    }

    override fun onMeasure(widthMeasureSpec: Int, heightMeasureSpec: Int) {
        super.onMeasure(widthMeasureSpec, heightMeasureSpec)
    }

    private fun Note.bindClickHandlers() {
        contentLayout.setOnClickListener { callbacks?.onItemClicked(this) }
    }

    private fun setContentLayoutBackgroundColor(color: Color) {
        setCardBackgroundColor(color.toContextColor(context))
    }

    protected fun loadFirstPhoto(noteImage: ImageView, note: Note) {
        note.document.mediaList().firstOrNull()?.let {
            noteImage.loadRoundedImageFromUri(it.getUrlOrNull(), 12F)
        }
    }

    protected fun archivedOrText(note: Note, nonArchivedText: String?): String? =
            if (note.isArchived) {
                context.formatString(R.string.label_archived_with_parameter, nonArchivedText)
            } else
                nonArchivedText

    open fun prepareSharedElements(markSharedElement: (View, String) -> Unit) {
        markSharedElement(this, "card")
        markSharedElement(contentLayout, "linearLayout")
    }

    open fun clearTransitionNames() {
        ViewCompat.setTransitionName(this, "")
        ViewCompat.setTransitionName(contentLayout, "")
    }

    fun setRootTransitionName(name: String) {
        clearTransitionNames()
        ViewCompat.setTransitionName(this, name)
    }

    open class Callbacks {
        open fun onItemClicked(note: Note) {}
        open fun onColorChanged(note: Note, color: Color) {}
        open fun onNotesReordered() {}
    }
}

class TextNoteItemComponent(context: Context, attrs: AttributeSet) : NoteItemComponent(context, attrs) {

    val noteTitle by lazy { findViewById<TextView>(R.id.noteTitle) }
    val noteBody by lazy { findViewById<TextView>(R.id.noteBody) }

    override fun bindNote(note: Note, shouldShowColorPicker: Boolean) {
        with(note) {
            val titleText = archivedOrText(this, title)
            noteTitle.setContentAndVisibility(titleText)

            val documentAsString = document.asString()
            val bodyText = archivedOrText(this, documentAsString)
            noteBody.setPreviewContentAndVisibility(bodyText, MAX_CHARACTER_COUNT_PREVIEW)
        }
        super.bindNote(note, shouldShowColorPicker)
    }

    override fun prepareSharedElements(markSharedElement: (View, String) -> Unit) {
        super.prepareSharedElements(markSharedElement)
        markSharedElement(noteTitle, "title")
        markSharedElement(noteBody, "body")
    }

    override fun clearTransitionNames() {
        super.clearTransitionNames()
        ViewCompat.setTransitionName(noteTitle, "")
        ViewCompat.setTransitionName(noteBody, "")
    }
}

class ImageNoteItemComponent(context: Context, attrs: AttributeSet) : NoteItemComponent(context, attrs) {

    val noteImage by lazy { findViewById<ImageView>(R.id.noteImage) }

    override fun bindNote(note: Note, shouldShowColorPicker: Boolean) {
        loadFirstPhoto(noteImage, note)
        super.bindNote(note, shouldShowColorPicker)
    }

    override fun prepareSharedElements(markSharedElement: (View, String) -> Unit) {
        super.prepareSharedElements(markSharedElement)
        markSharedElement(noteImage, "image")
    }

    override fun clearTransitionNames() {
        super.clearTransitionNames()
        ViewCompat.setTransitionName(noteImage, "")
    }
}

class MixedNoteItemComponent(context: Context, attrs: AttributeSet) : NoteItemComponent(context, attrs) {

    val noteTitle by lazy { findViewById<TextView>(R.id.noteTitle) }
    val noteBody by lazy { findViewById<TextView>(R.id.noteBody) }
    val noteImage by lazy { findViewById<ImageView>(R.id.noteImage) }

    override fun bindNote(note: Note, shouldShowColorPicker: Boolean) {
        with(note) {
            val titleText = archivedOrText(this, title)
            noteTitle.setContentAndVisibility(titleText)

            val documentAsString = document.asString()
            val bodyText = archivedOrText(this, documentAsString)
            noteBody.setPreviewContentAndVisibility(bodyText, MAX_CHARACTER_COUNT_PREVIEW)

            loadFirstPhoto(noteImage, note)
        }
        super.bindNote(note, shouldShowColorPicker)
    }

    override fun prepareSharedElements(markSharedElement: (View, String) -> Unit) {
        super.prepareSharedElements(markSharedElement)
        markSharedElement(noteTitle, "title")
        markSharedElement(noteBody, "body")
        markSharedElement(noteImage, "image")
    }

    override fun clearTransitionNames() {
        super.clearTransitionNames()
        ViewCompat.setTransitionName(noteTitle, "")
        ViewCompat.setTransitionName(noteBody, "")
        ViewCompat.setTransitionName(noteImage, "")
    }
}

class WrongViewHolderTypeException : ClassNotFoundException() {
    override val message: String?
        get() = "Wrong view holder!!"
}
