package com.microsoft.notes.noteslist.recyclerview

import android.support.v4.view.ViewCompat
import android.support.v7.util.DiffUtil
import android.support.v7.widget.RecyclerView
import android.view.LayoutInflater
import android.view.ViewGroup
import com.microsoft.notes.R
import com.microsoft.notes.extensions.getStableId
import com.microsoft.notes.store.Color
import com.microsoft.notes.store.Note
import com.microsoft.notes.store.extensions.findAndMap
import java.util.Collections

class NotesListAdapter(
        private var notes: List<Note>,
        private val getExpandedPosition: () -> Int,
        private val itemClick: (Note) -> Unit,
        private val setColor: (Note, Color) -> Unit,
        private val updateNotesPositions: (List<Note>) -> Unit)
    : RecyclerView.Adapter<NoteViewHolder>(), ItemTouchHelperAdapter {

    var noteWithColorPickerOpen: Note? = null

    private val itemCallbacks = object : NoteItemComponent.Callbacks() {
        override fun onItemClicked(note: Note) {
            itemClick(note)
        }

        override fun onColorChanged(note: Note, color: Color) {
            setColor(note, color)
        }

        override fun onNotesReordered() {
            updateNotesPositions()
        }
    }

    enum class NoteItemType(val id: Int) {
        TEXT(0), IMAGE(1), MIXED(2)
    }

    init {
        setHasStableIds(true)
    }

    override fun onCreateViewHolder(parent: ViewGroup?, viewType: Int): NoteViewHolder {
        val layoutId = when (viewType) {
            NoteItemType.TEXT.id -> R.layout.note_item_layout_text
            NoteItemType.IMAGE.id -> R.layout.note_item_layout_image
            NoteItemType.MIXED.id -> R.layout.note_item_layout_mixed
            else -> throw WrongViewHolderTypeException()
        }

        val view = LayoutInflater.from(parent?.context).inflate(layoutId, parent, false) as NoteItemComponent
        view.callbacks = itemCallbacks
        return NoteViewHolder(view)
    }

    override fun onBindViewHolder(holderNote: NoteViewHolder, position: Int) {
        with(holderNote.noteView) {
            bindNote(notes[position], notes[position] == noteWithColorPickerOpen)
            if (position == getExpandedPosition()) {
                prepareSharedElements(ViewCompat::setTransitionName)
            } else if (position < getExpandedPosition()) {
                setRootTransitionName("up")
            } else if (position > getExpandedPosition()) {
                setRootTransitionName("down")
            }
        }
    }

    override fun getItemViewType(position: Int): Int {
        with(notes[position]) {
            if (isMediaListEmpty) return NoteItemType.TEXT.id
            else if (hasNoText) return NoteItemType.IMAGE.id
            else return NoteItemType.MIXED.id
        }
    }

    override fun getItemCount(): Int = notes.size

    fun getItem(position: Int): Note = notes[position]

    override fun getItemId(position: Int): Long =
            getItem(position).getStableId()

    fun removeAt(position: Int) {
        notes = notes.minus(notes[position])
        notifyItemRemoved(position)
    }

    fun updateNotesCollection(notesCollection: List<Note>) {
        val oldNotes = notes
        notes = notesCollection
        applyDiff(oldNotes, notes)
    }

    private fun applyDiff(oldNotes: List<Note>, newNotesCollection: List<Note>) {
        val diffResult = DiffUtil.calculateDiff(NotesDiffCallback(oldNotes, newNotesCollection))
        diffResult.dispatchUpdatesTo(this)
    }

    private fun updateNotesPositions() {
        updateNotesPositions(notes)
    }

    override fun onItemMove(fromPosition: Int, toPosition: Int) {
        swapItems(fromPosition, toPosition)
        notifyItemMoved(fromPosition, toPosition)
    }

    fun swapItems(fromPosition: Int, toPosition: Int) = if (fromPosition < toPosition) {
        (fromPosition .. toPosition - 1).forEach { i ->
            swapPositions(i, i + 1)
            Collections.swap(notes, i, i + 1)
        }
    } else {
        (fromPosition downTo toPosition + 1).forEach { i ->
            swapPositions(i, i - 1)
            Collections.swap(notes, i, i - 1)
        }
    }

    fun swapPositions(position1: Int, position2: Int) {
        val item1 = notes[position1]
        val item2 = notes[position2]
        notes = notes.findAndMap(
                find = { it.localId == item1.localId }, map = { it.copy(position = item2.position) }
        ).findAndMap(
                find = { it.localId == item2.localId }, map = { it.copy(position = item1.position) })
    }
}