package com.microsoft.notes.photos

import android.Manifest
import android.Manifest.permission.READ_EXTERNAL_STORAGE
import android.app.Activity
import android.content.ComponentName
import android.content.Context
import android.content.Intent
import android.content.pm.ActivityInfo
import android.content.pm.PackageManager.PERMISSION_GRANTED
import android.content.pm.ResolveInfo
import android.os.AsyncTask
import android.os.Environment
import android.provider.MediaStore
import android.support.v4.app.ActivityCompat
import android.support.v4.content.ContextCompat.checkSelfPermission
import android.support.v4.content.FileProvider
import com.microsoft.notes.R
import com.microsoft.notes.components.ActivityResult
import com.microsoft.notes.components.NotesViewNavigation.REQUEST_CAMERA_PERMISSION
import com.microsoft.notes.components.NotesViewNavigation.REQUEST_PICK_PHOTO
import com.microsoft.notes.components.NotesViewNavigation.REQUEST_READ_EXTERNAL_STORAGE_PERMISSION
import com.microsoft.notes.components.NotesViewNavigation.REQUEST_TAKE_PHOTO
import com.microsoft.notes.components.ViewActivityComponent
import com.microsoft.notes.extensions.getLocale
import rx.Observable
import rx.android.schedulers.AndroidSchedulers
import rx.schedulers.Schedulers
import java.io.File
import java.text.SimpleDateFormat
import java.util.Date

// Bug 2057090: Move Content provider out of UI package and use authority from there
val ONENOTE_FILE_PROVIDER_AUTHORITY = "com.microsoft.onenote.file.provider"

enum class CameraSlideInDirection {
    LEFT,
    RIGHT,
    NONE
}

internal fun createImageFileAsync(activity: Activity): Observable<File> {
    // Create an image file name
    val timeStamp = SimpleDateFormat("yyyyMMdd_HHmmss", activity.getLocale()).format(Date())
    val imageFileName = timeStamp + "_"
    val storageDir = activity.getExternalFilesDir(Environment.DIRECTORY_PICTURES)
    return Observable.defer {
        Observable.just(File.createTempFile(
                imageFileName, /* prefix */
                ".png", /* suffix */
                storageDir      /* directory */))
    }.subscribeOn(Schedulers.io()).observeOn(AndroidSchedulers.mainThread())
}

internal fun File.deleteFileAsync() {
    AsyncTask.execute { delete() }
}

internal fun takePhoto(activityComponent: ViewActivityComponent,
                       file: File,
                       packageName: String,
                       activityName: String,
                       cameraSlideInDirection: CameraSlideInDirection,
                       permissionDeniedCallback: () -> Unit,
                       requestCompletedCallback: (ActivityResult) -> Unit) {
    with(activityComponent) {
        if (checkSelfPermission(this.activity, Manifest.permission.CAMERA) != PERMISSION_GRANTED) {
            ActivityCompat.requestPermissions(this.activity, arrayOf(Manifest.permission.CAMERA),
                    REQUEST_CAMERA_PERMISSION)
            subActivityResultSubscribers.permissionRequestResults.first { it.requestCode == REQUEST_CAMERA_PERMISSION }
                    .observeOn(AndroidSchedulers.mainThread())
                    .subscribe {
                        if (it.grantResults.isNotEmpty() && it.grantResults[0] == PERMISSION_GRANTED) {
                            doTakePhoto(activityComponent, file, packageName, activityName, cameraSlideInDirection, requestCompletedCallback)
                        } else {
                            permissionDeniedCallback()
                        }
                    }
        } else {
            doTakePhoto(activityComponent, file, packageName, activityName, cameraSlideInDirection, requestCompletedCallback)
        }
    }
}

internal fun doTakePhoto(activityComponent: ViewActivityComponent,
                         file: File,
                         packageName: String,
                         activityName: String,
                         cameraSlideInDirection: CameraSlideInDirection,
                         requestCompletedCallback: (ActivityResult) -> Unit) {
    with(activityComponent) {
        val intent = createTakePhotoIntent()
        intent.component = ComponentName(packageName, activityName)
        if (intent.resolveActivity(activity.packageManager) != null) {
            val photoURI = FileProvider.getUriForFile(activity, ONENOTE_FILE_PROVIDER_AUTHORITY, file)
            intent.putExtra(MediaStore.EXTRA_OUTPUT, photoURI)
            activity.grantUriPermission(packageName, photoURI,
                    Intent.FLAG_GRANT_WRITE_URI_PERMISSION or Intent.FLAG_GRANT_READ_URI_PERMISSION)
            activity.startActivityForResult(intent, REQUEST_TAKE_PHOTO)
            when (cameraSlideInDirection) {
                CameraSlideInDirection.LEFT -> activity.overridePendingTransition(R.animator.slide_from_left, R.animator.slide_to_right)
                CameraSlideInDirection.RIGHT -> activity.overridePendingTransition(R.animator.slide_from_right, R.animator.slide_to_left)
                else -> { /* Do nothing */ }
            }
            subActivityResultSubscribers.activityResults.first { it.requestCode == REQUEST_TAKE_PHOTO }
                    .observeOn(AndroidSchedulers.mainThread())
                    .subscribe { requestCompletedCallback(it) }
        }
    }
}

internal fun pickPhoto(activityComponent: ViewActivityComponent,
                       packageName: String?,
                       activityName: String?,
                       permissionDeniedCallback: () -> Unit,
                       requestCompletedCallback: (ActivityResult) -> Unit) {
    with(activityComponent) {
        if (checkSelfPermission(activity, READ_EXTERNAL_STORAGE) != PERMISSION_GRANTED) {
            ActivityCompat.requestPermissions(activity, arrayOf(READ_EXTERNAL_STORAGE),
                    REQUEST_READ_EXTERNAL_STORAGE_PERMISSION)
            subActivityResultSubscribers.permissionRequestResults.first {
                it.requestCode ==
                        REQUEST_READ_EXTERNAL_STORAGE_PERMISSION
            }
                    .observeOn(AndroidSchedulers.mainThread())
                    .subscribe {
                        if (it.grantResults.isNotEmpty() && it.grantResults[0] == PERMISSION_GRANTED) {
                            doPickPhoto(packageName, activityName, requestCompletedCallback)
                        } else {
                            permissionDeniedCallback()
                        }
                    }
        } else {
            doPickPhoto(packageName, activityName, requestCompletedCallback)
        }
    }
}

private fun ViewActivityComponent.doPickPhoto(packageName: String?, activityName: String?,
                                              requestCompletedCallback: (ActivityResult) -> Unit) {
    val intent = createPickPhotoIntent()
    if (packageName != null && activityName != null)
        intent.component = ComponentName(packageName, activityName)
    if (intent.resolveActivity(activity.packageManager) != null) {
        activity.startActivityForResult(intent, REQUEST_PICK_PHOTO)
        subActivityResultSubscribers.activityResults.first { it.requestCode == REQUEST_PICK_PHOTO }
                .observeOn(AndroidSchedulers.mainThread())
                .subscribe { requestCompletedCallback(it) }
    }
}

fun createTakePhotoIntent() = Intent(MediaStore.ACTION_IMAGE_CAPTURE)

fun getDefaultCameraActivityInfo(context: Context) : ActivityInfo? {
    val takePictureResults: MutableList<ResolveInfo>? =
            context.packageManager.queryIntentActivities(createTakePhotoIntent(), 0)
    if(takePictureResults?.isNotEmpty() ?: false) {
        return takePictureResults!![0].activityInfo
    }
    return null
}

fun createPickPhotoIntent(): Intent {
    val intent: Intent = Intent(Intent.ACTION_PICK, MediaStore.Images.Media.EXTERNAL_CONTENT_URI)
    intent.type = "image/*"
    return intent
}