package com.microsoft.notes.search

import android.os.Build
import android.os.Bundle
import android.os.Handler
import android.os.HandlerThread
import android.os.Looper
import android.view.LayoutInflater
import android.view.Menu
import android.view.MenuInflater
import android.view.View
import android.view.ViewGroup
import android.widget.SearchView
import com.microsoft.notes.R
import com.microsoft.notes.ViewFragment
import com.microsoft.notes.appstore.Navigation
import com.microsoft.notes.appstore.ScrollTo
import com.microsoft.notes.appstore.getCurrentEditNote
import com.microsoft.notes.components.UiNotesFragmentComponent
import com.microsoft.notes.components.ViewActivityWithStateManagerComponent
import com.microsoft.notes.extensions.replaceChild
import com.microsoft.notes.extensions.setupToolbar
import com.microsoft.notes.noteslist.NotesListComponent
import com.microsoft.notes.noteslist.recyclerview.NoteViewHolder
import com.microsoft.notes.store.Color
import com.microsoft.notes.store.Note
import com.microsoft.notes.utils.logging.EventMarkers
import com.microsoft.notes.utils.logging.NotesOperation
import com.microsoft.notes.utils.logging.TelemetryKeys
import com.microsoft.notes.transition.SearchTransitions
import kotlinx.android.synthetic.main.search_fragment_layout.*

class SearchFragment : ViewFragment<SearchControllerView>(), UiNotesFragmentComponent, Search {

    init {
        allowEnterTransitionOverlap = true
        allowReturnTransitionOverlap = true
    }

    val notesList get() = view?.findViewById<NotesListComponent>(R.id.notesList)
    
    private val mainHandler = Handler(Looper.getMainLooper())

    override fun onCreateView(inflater: LayoutInflater, container: ViewGroup?, savedInstanceState: Bundle?): View {
        return inflater.inflate(R.layout.search_fragment_layout, container, false)
    }

    override fun onActivityCreated(savedInstanceState: Bundle?) {
        super.onActivityCreated(savedInstanceState)

        setupToolbar()
        setupSearchView()
        updateNotesListVisibility(false)
    }

    override fun searchAsync(asyncThreadBlock: () -> List<Note>, uiThreadBlock: (List<Note>, ScrollTo) -> Unit) {
        val asyncThread = HandlerThread("AsyncThread")
        asyncThread.start()
        val handler = Handler(asyncThread.looper)
        handler.post({
            val result = asyncThreadBlock()
            mainHandler.post {
                uiThreadBlock(result, ScrollTo.Top)
            }
        })
    }

    override fun setupControllerView() = SearchControllerView(uiFragmentComponent = this, search = this)

    private fun setupToolbar() {
        parent().setupToolbar(notesToolbar)
        toolBarBackButton.setOnClickListener {
            activity.onBackPressed()
        }
        setHasOptionsMenu(true)
    }

    private fun setupNotesList() {
        notesList!!.callbacks = object : NotesListComponent.Callbacks() {

            override val noteForReturnTransition: Note?
                get() = cv?.appState?.getCurrentEditNote()

            override fun onNoteColorChanged(note: Note, color: Color) {
                recordNotesOperation(NotesOperation.NoteColorChanged)
                cv?.setNoteColor(note, color)                
            }

            override fun onNoteClicked(note: Note) {
                recordNotesOperation(NotesOperation.NoteClicked)
                editNoteWithContent(note)
            }
        }
    }

    private fun recordNotesOperation(notesOperation: NotesOperation) {
        cv?.recordTelemetry(EventMarkers.NotesOperationTriggered,
                Pair(TelemetryKeys.NOTES_OPERATION, notesOperation.name),
                Pair(TelemetryKeys.TRIGGER_POINT, Navigation.SEARCH.name))
    }

    private fun setupSearchView() {
        searchView.onActionViewExpanded()
        searchView.setOnQueryTextListener(object : SearchView.OnQueryTextListener {
            override fun onQueryTextSubmit(query: String?): Boolean = true

            override fun onQueryTextChange(newText: String): Boolean {
                cv?.query = newText
                return true
            }
        })
    }

    override fun onCreateOptionsMenu(menu: Menu?, inflater: MenuInflater?) {
        super.onCreateOptionsMenu(menu, inflater)

        menu?.findItem(R.id.actionSearch)?.isVisible = false
        menu?.findItem(R.id.actionNotebooksMode)?.isVisible = false
        menu?.findItem(R.id.options_notes_about)?.isVisible = false
    }

    override fun parent(): ViewActivityWithStateManagerComponent = activity as ViewActivityWithStateManagerComponent

    fun findCurrentNoteView(): NoteViewHolder? =
            cv?.appState?.getCurrentEditNote()?.let { notesList?.findNoteView(it) }

    private fun editNoteWithContent(note: Note) {
        cv?.navigatingToNote(note)
    }

    private fun updateNotesListVisibility(animate: Boolean) {
        if (cv?.query?.isNotEmpty()?:false && notesList == null) {
            addNotesList(animate)
        }
    }

    private fun addNotesList(animate: Boolean) {
        contentContainer.replaceChild(R.layout.search_results_layout,
                if (animate && Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP)
                    SearchTransitions.colorBubblesToResults(context) else null)
        setupNotesList()
    }

    override fun updateNotesCollection(notesCollection: List<Note>, scrollTo: ScrollTo) {
        updateNotesListVisibility(true)
        notesList?.updateNotesCollection(notesCollection, scrollTo)
    }
}