package com.microsoft.notes.transition

import android.graphics.Path
import android.os.Build
import android.support.annotation.RequiresApi
import android.transition.PathMotion

// Copied from Android SDK 25 rev. 3, with changes on lines 54 and 67 to fix a bug in the SDK
@RequiresApi(Build.VERSION_CODES.LOLLIPOP)
class ArcMotion : PathMotion() {
    private val DEFAULT_MAX_ANGLE_DEGREES = 70f
    private val DEFAULT_MAX_TANGENT = Math.tan(Math.toRadians((DEFAULT_MAX_ANGLE_DEGREES / 2).toDouble())).toFloat()

    private var minimumHorizontalTangent = 0f
    private var minimumVerticalTangent = 0f
    private var maximumTangent = DEFAULT_MAX_TANGENT

    override fun getPath(startX: Float, startY: Float, endX: Float, endY: Float): Path {
        val path = Path()
        path.moveTo(startX, startY)

        var ex: Float
        var ey: Float
        if (startY == endY) {
            ex = (startX + endX) / 2
            ey = startY + minimumHorizontalTangent * Math.abs(endX - startX) / 2
        } else if (startX == endX) {
            ex = startX + minimumVerticalTangent * Math.abs(endY - startY) / 2
            ey = (startY + endY) / 2
        } else {
            val deltaX = endX - startX
            val deltaY = endY - startY
            // hypotenuse squared.
            val h2 = deltaX * deltaX + deltaY * deltaY

            // Midpoint between start and end
            val dx = (startX + endX) / 2
            val dy = (startY + endY) / 2

            // Distance squared between end point and mid point is (1/2 hypotenuse)^2
            val midDist2 = h2 * 0.25f

            var minimumArcDist2 = 0f

            if (Math.abs(deltaX) < Math.abs(deltaY)) {
                // Similar triangles bfa and bde mean that (ab/fb = eb/bd)
                // Therefore, eb = ab * bd / fb
                // ab = hypotenuse
                // bd = hypotenuse/2
                // fb = deltaY
                val eDistY = h2 / (2 * deltaY)
                // CHANGED FROM ANDROID SDK: Prefer motion that follows gravity, per Material spec
                if (deltaY > 0) {
                    ey = startY + eDistY
                    ex = startX
                } else {
                    ey = endY - eDistY
                    ex = endX
                }

                minimumArcDist2 = midDist2 * minimumVerticalTangent * minimumVerticalTangent
            } else {
                // Same as above, but flip X & Y
                val eDistX = h2 / (2 * deltaX)
                // CHANGED FROM ANDROID SDK: Prefer motion that follows gravity, per Material spec
                if (deltaY > 0) {
                    ex = endX - eDistX
                    ey = endY
                } else {
                    ex = startX + eDistX
                    ey = startY
                }

                minimumArcDist2 = midDist2 * minimumHorizontalTangent * minimumHorizontalTangent
            }
            val arcDistX = dx - ex
            val arcDistY = dy - ey
            val arcDist2 = arcDistX * arcDistX + arcDistY * arcDistY

            val maximumArcDist2 = midDist2 * maximumTangent * maximumTangent

            var newArcDistance2 = 0f
            if (arcDist2 < minimumArcDist2) {
                newArcDistance2 = minimumArcDist2
            } else if (arcDist2 > maximumArcDist2) {
                newArcDistance2 = maximumArcDist2
            }
            if (newArcDistance2 != 0f) {
                val ratio2 = newArcDistance2 / arcDist2
                val ratio = Math.sqrt(ratio2.toDouble()).toFloat()
                ex = dx + ratio * (ex - dx)
                ey = dy + ratio * (ey - dy)
            }
        }
        val controlX1 = (startX + ex) / 2
        val controlY1 = (startY + ey) / 2
        val controlX2 = (ex + endX) / 2
        val controlY2 = (ey + endY) / 2
        path.cubicTo(controlX1, controlY1, controlX2, controlY2, endX, endY)
        return path
    }
}