package com.microsoft.notes.transition

import android.animation.ObjectAnimator
import android.animation.TimeInterpolator
import android.content.Context
import android.graphics.Bitmap
import android.graphics.Canvas
import android.graphics.Rect
import android.graphics.drawable.BitmapDrawable
import android.os.Build
import android.support.annotation.RequiresApi
import android.support.v4.app.Fragment
import android.support.v4.app.FragmentActivity
import android.transition.Transition
import android.transition.TransitionSet
import android.view.Gravity
import android.view.View
import android.view.ViewGroup
import android.widget.ImageView
import com.microsoft.notes.R
import com.microsoft.notes.editnote.EditNoteFragment
import com.microsoft.notes.extensions.accelerateAndDecelerate
import com.microsoft.notes.extensions.addTargets
import com.microsoft.notes.extensions.addTransitions
import com.microsoft.notes.extensions.decelerate
import com.microsoft.notes.extensions.getBoundsOnScreen
import com.microsoft.notes.extensions.getLocationOnScreenAsPoint
import com.microsoft.notes.extensions.nextLayoutListener
import com.microsoft.notes.extensions.onEnd
import com.microsoft.notes.extensions.setLayoutParamsFrom
import com.microsoft.notes.extensions.withSidePropagation
import com.microsoft.notes.noteslist.NotesListFragment
import com.microsoft.notes.noteslist.recyclerview.NoteItemComponent
import kotlinx.android.synthetic.main.edit_note_content_layout.*
import kotlinx.android.synthetic.main.notes_list_component_layout.*

@RequiresApi(Build.VERSION_CODES.LOLLIPOP)
object NotesListEditNoteTransitions {

    const val MIN_DURATION = 150L
    const val EXTRA_DURATION = 200L
    const val TOOLBAR_DURATION = 200L
    const val NEW_NOTE_ENTER_DURATION = 400L
    const val SLIDE_NOTES_UNDER_DURATION = 267L

    fun applyToEditNote(activity: FragmentActivity, from: Fragment, to: EditNoteFragment, isNewNote: Boolean,
                        animationOverlay: ViewGroup) {
        from.exitTransition = notesListEnterExit(activity, from, isNewNote, false)
        from.enterTransition = null
        to.enterTransition = editNoteEnterExit(activity, isEntering = true, isEmpty = false,
                animateBackButton = from is NotesListFragment)
        to.exitTransition = null
        to.sharedElementEnterTransition = sharedElements(true)

        if (isNewNote) {
            animateEditViewForNewNote(activity, to, animationOverlay)
        }
    }

    fun applyFromEditNote(context: Context, from: EditNoteFragment, to: Fragment, isEmpty: Boolean) {
        from.exitTransition = editNoteEnterExit(context, isEntering = false, isEmpty = isEmpty,
                animateBackButton = to is NotesListFragment)
        from.enterTransition = null
        to.enterTransition = notesListEnterExit(context, to, false, true)
        to.exitTransition = null
        to.sharedElementEnterTransition = sharedElements(false)
    }

    object params : DynamicTransitionParams.Duration, DynamicTransitionParams.VerticalSlideDistance {
        override var duration: Long = 0L
        override var slideUpDistance: Int = 0
        override var slideDownDistance: Int = 0
        override var containerBounds: Rect = Rect()

        fun setFromViews(noteView: View, noteListView: View) {
            val notePos = IntArray(2)
            val noteListPos = IntArray(2)
            noteView.getLocationOnScreen(notePos)
            noteListView.getLocationOnScreen(noteListPos)

            slideUpDistance = noteListPos[1] - notePos[1]
            slideDownDistance = noteListPos[1] + noteListView.height - (notePos[1] + noteView.height)
            containerBounds = Rect(noteListPos[0], noteListPos[1], noteListPos[0] + noteListView.width,
                    noteListPos[1] + noteListView.height)
            val magnitude: Float = maxOf(Math.abs(slideUpDistance),
                    slideDownDistance).toFloat() / noteListView.height
            duration = Math.round(MIN_DURATION + maxOf(magnitude, 0f) * EXTRA_DURATION).toLong()
        }
    }

    private fun notesListEnterExit(context: Context, notesView: Fragment, isNewNote: Boolean, isEntering: Boolean):
            TransitionSet {
        val toolbarSlide = context.resources.getDimension(R.dimen.toolbar_translate_animation_negative)
        return TransitionSet().addTransitions(
                slideAndFadeToolbarTitle(toolbarSlide),
                fadeSearchButtonAndView(),
                when {
                    isNewNote && !isEntering -> slideNotesUnder(notesView)
                    else -> slideNotesInOrOut()
                }
        )
    }

    private fun slideAndFadeToolbarTitle(toolbarSlide: Float): Transition =
            ScaleTranslateAndFadeTransition(params, dx = toolbarSlide).apply {
                addTarget(R.id.toolBarTitle)
                accelerateAndDecelerate()
            }

    private fun fadeSearchButtonAndView(): Transition =
            Fade().apply {
                addTargets(R.id.searchView, R.id.actionSearch)
                duration = TOOLBAR_DURATION
                accelerateAndDecelerate()
            }

    private fun slideNotesUnder(notesView: Fragment): Transition {
        val recyclerViewBounds = notesView.notesRecyclerView.getBoundsOnScreen()
        return ScaleTranslateAndFadeTransition(dy = 70f, scale = .85f, clipBounds = recyclerViewBounds).apply {
            addTarget(NoteItemComponent::class.java)
            duration = SLIDE_NOTES_UNDER_DURATION
            withSidePropagation(Gravity.BOTTOM, 2f)
            accelerateAndDecelerate()
        }
    }

    private fun slideNotesInOrOut(): Transition =
            DynamicDurationTransitionSet().addTransitions(
                    VerticalSlideWithCompression(Gravity.TOP, params, params).addTarget("up"),
                    VerticalSlideWithCompression(Gravity.BOTTOM, params, params).addTarget("down")
            )

    private fun sharedElements(toEditNote: Boolean): DynamicDurationTransitionSet {
        with(DynamicDurationTransitionSet()) {
            addTransitions(
                    ChangeBounds(params).addTargets("card", "linearLayout", "colorLayout", "title", "body"),
                    TitleTextTransition(params, toEditNote).addTarget("title"),
                    BodyTextTransition(params, toEditNote).addTarget("body"),
                    CardElevationTransition(params).addTarget("card")
            )

            return this
        }
    }

    private fun editNoteEnterExit(context: Context, isEntering: Boolean,
                                  isEmpty: Boolean, animateBackButton: Boolean): TransitionSet =
            TransitionSet().apply {
                if (!isEntering && isEmpty)
                    addTransition(fadeEmptyNoteOnExit())

                if (animateBackButton)
                    addTransition(slideAndFadeBackButton(context))
            }

    private fun fadeEmptyNoteOnExit(): Transition =
            Fade(params).addTarget("card").decelerate()

    private fun slideAndFadeBackButton(context: Context): Transition {
        val toolbarSlide = context.resources.getDimension(R.dimen.toolbar_translate_animation_positive)
        return ScaleTranslateAndFadeTransition(dx = toolbarSlide)
                .addTarget(R.id.toolBarBackButton)
                .setDuration(TOOLBAR_DURATION)
                .accelerateAndDecelerate()
    }

    private fun animateEditViewForNewNote(activity: FragmentActivity, editNoteFragment: EditNoteFragment,
                                          animationOverlay: ViewGroup) {
        editNoteFragment.addFirstLayoutListener {
            val editNoteOverlay = createEditNoteOverlay(activity, editNoteFragment, animationOverlay)
            editNoteFragment.editNoteCardView.visibility = View.INVISIBLE

            animationOverlay.nextLayoutListener {
                editNoteOverlay.translationY = editNoteOverlay.height.toFloat()
                moveEditNoteAnimator(editNoteOverlay)
                        .onEnd {
                            animationOverlay.removeAllViews()
                            if(editNoteFragment.isAdded) {
                                editNoteFragment.editNoteCardView.visibility = View.VISIBLE
                                editNoteFragment.onNavigateToTransitionCompleted()
                            }
                        }.start()
            }
        }
    }

    private fun createEditNoteOverlay(activity: FragmentActivity, editNoteFragment: EditNoteFragment,
                                      animationOverlay: ViewGroup):
            ImageView {
        val cardView = editNoteFragment.editNoteCardView
        val bitmap = Bitmap.createBitmap(cardView.measuredWidth, cardView.measuredHeight, Bitmap.Config.ARGB_8888)
        val canvas = Canvas(bitmap)
        cardView.draw(canvas)
        val overlayPosition = animationOverlay.getLocationOnScreenAsPoint()
        val imageView = ImageView(activity)
        imageView.setImageDrawable(BitmapDrawable(activity.resources, bitmap))
        animationOverlay.addView(imageView)
        imageView.setLayoutParamsFrom(editNoteFragment.editNoteCardView, overlayPosition)
        return imageView
    }

    private fun moveEditNoteAnimator(view: View) =
            ObjectAnimator.ofFloat(view, View.TRANSLATION_Y, view.height.toFloat(), 0f).apply {
                duration = NEW_NOTE_ENTER_DURATION
                accelerateAndDecelerate()
            }

    private open class DynamicDurationTransitionSet : TransitionSet() {
        init {
            accelerateAndDecelerate()
            ordering = TransitionSet.ORDERING_TOGETHER
        }

        override fun getDuration(): Long = params.duration

        override fun addTransition(transition: Transition): TransitionSet {
            if (interpolator != null)
                transition.interpolator = interpolator

            return super.addTransition(transition)
        }

        override fun setInterpolator(interpolator: TimeInterpolator?): TransitionSet {
            super.setInterpolator(interpolator)

            if (interpolator != null) {
                val numTransitions = transitionCount
                for (i in 0 .. numTransitions - 1) {
                    getTransitionAt(i).interpolator = interpolator
                }
            }

            return this
        }
    }
}