package com.microsoft.notes.transition

import android.animation.AnimatorSet
import android.animation.ObjectAnimator
import android.animation.PropertyValuesHolder
import android.content.Context
import android.graphics.Point
import android.os.Build
import android.support.annotation.RequiresApi
import android.support.v4.app.FragmentActivity
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import com.microsoft.notes.R
import com.microsoft.notes.extensions.accelerate
import com.microsoft.notes.extensions.accelerateOrDecelerate
import com.microsoft.notes.extensions.addTargets
import com.microsoft.notes.extensions.animateSequentially
import com.microsoft.notes.extensions.animateTogether
import com.microsoft.notes.extensions.childrenSequence
import com.microsoft.notes.extensions.decelerate
import com.microsoft.notes.extensions.getBoundsOnScreen
import com.microsoft.notes.extensions.getLocationOnScreenAsPoint
import com.microsoft.notes.extensions.minus
import com.microsoft.notes.extensions.onEnd
import com.microsoft.notes.extensions.plus
import com.microsoft.notes.extensions.setLayoutParamsFrom
import com.microsoft.notes.noteslist.NotesListComponent
import com.microsoft.notes.noteslist.NotesListFragment
import com.microsoft.notes.noteslist.recyclerview.NoteItemComponent
import com.microsoft.notes.noteslist.recyclerview.NoteViewHolder
import com.microsoft.notes.search.SearchFragment
import com.microsoft.notes.store.Note
import kotlinx.android.synthetic.main.notes_layout.*
import kotlinx.android.synthetic.main.notes_list_component_layout.view.*

@RequiresApi(Build.VERSION_CODES.LOLLIPOP)
object NotesListToSearchTransitions {
    private val SEARCH_CONTENT_FADE_DURATION = 150L
    private val NOTES_LIST_FADE_DURATION = 150L
    private val RESULTS_TO_NOTES_LIST_MOVE_DURATION = 200L
    private val RESULTS_TO_NOTES_LIST_FADE_DURATION = 200L
    private val RESULTS_TO_NOTES_LIST_FADE_DELAY = 25L
    private fun searchContentFadeDelay(entering: Boolean) = if (entering) 75L else 0L
    private fun notesListFadeDelay(entering: Boolean) = if (entering) 75L else 0L

    fun apply(activity: FragmentActivity, from: NotesListFragment, to: SearchFragment) {
        from.exitTransition = notesViewEnterExit(activity, entering = false, animateNotes = true)
        from.enterTransition = null
        to.enterTransition = searchEnterExit(activity, true)
        to.exitTransition = null
        to.sharedElementEnterTransition = null
    }

    fun apply(activity: FragmentActivity, from: SearchFragment, to: NotesListFragment,
              animationOverlay: ViewGroup) {
        from.enterTransition = null
        from.exitTransition = searchEnterExit(activity, false)
        to.enterTransition = notesViewEnterExit(activity, entering = true, animateNotes = true)
        to.exitTransition = null
        to.sharedElementEnterTransition = null

        if(from.notesList != null && ((from.notesList as? NotesListComponent)?.notesRecyclerView?.childrenSequence()?.toList()?.isNotEmpty() ?: false))
        {
            searchResultsToNotesList(activity, from, to, animationOverlay)
        }
    }

    private fun searchEnterExit(context: Context, entering: Boolean) =
            ScaleTranslateAndFadeTransition(
                    dy = -context.resources.getDimension(R.dimen.color_filter_translate_animation))
                    .apply {
                        addTargets(R.id.searchView, R.id.toolBarBackButton, R.id.contentContainer)
                        duration = SEARCH_CONTENT_FADE_DURATION
                        startDelay = searchContentFadeDelay(entering)
                        accelerateOrDecelerate(!entering)
                    }

    private fun notesViewEnterExit(context: Context, entering: Boolean, animateNotes: Boolean) =
            when {
                entering -> slideAndFadeIn(context)
                else -> shrinkAndSlideOut(context)
            }.apply {
                addTargets(R.id.toolBarTitle, R.id.actionSearch, R.id.newNoteFab)
                if (animateNotes)
                    addTarget(NoteItemComponent::class.java)
                duration = NOTES_LIST_FADE_DURATION
                startDelay = notesListFadeDelay(entering)
            }

    private fun searchResultsToNotesList(activity: FragmentActivity, from: SearchFragment, to: NotesListFragment,
                                         animationOverlay: ViewGroup) {
        val overlayPosition = animationOverlay.getLocationOnScreenAsPoint()
        val noteOverlays = createResultOverlays(from, activity, overlayPosition, animationOverlay)
        from.notesList!!.visibility = View.GONE

        to.addFirstLayoutListener {
            if(to.notesList != null && ((to.notesList as? NotesListComponent)?.notesRecyclerView?.childrenSequence()?.toList()?.isNotEmpty() ?: false))
            {
                clipOverlayToView(activity, to.notesList, overlayPosition, animationOverlay)
                to.notesList.alpha = 0f

                animateSequentially(
                        moveResultOverlays(activity, noteOverlays, to, overlayPosition, animationOverlay),
                        animateNotesListFadeAfterResultsMove(to.notesList)
                ).onEnd {
                    animationOverlay.removeAllViews()
                }.start()
            }
        }
    }

    private fun moveResultOverlays(activity: FragmentActivity, noteOverlays: List<NoteItemComponent>,
                                   to: NotesListFragment, overlayPosition: Point,
                                   animationOverlay: ViewGroup): AnimatorSet {
        val toRecyclerView = to.notesList.notesRecyclerView
        val targetViews = toRecyclerView.childrenSequence().toList().map {
            toRecyclerView.findContainingViewHolder(it) as NoteViewHolder
        }.associateBy {
            it.sourceNote.localId
        }

        val firstTargetViewIndex = targetViews.values.map { it.adapterPosition }.min()!!
        val overlayHeight = animationOverlay.height
        val moveOverlays = noteOverlays.map { overlay ->
            animateResultOverlay(overlay, targetViews, to, firstTargetViewIndex, overlayHeight, overlayPosition)
        }.animateTogether().apply {
            duration = RESULTS_TO_NOTES_LIST_MOVE_DURATION
            decelerate()
        }
        return moveOverlays
    }

    private fun animateResultOverlay(overlay: NoteItemComponent, targetViews: Map<String, NoteViewHolder>,
                                     to: NotesListFragment, firstTargetViewIndex: Int, overlayHeight: Int,
                                     overlayPosition: Point): ObjectAnimator {
        val note = overlay.sourceNote
        val target = targetViews[overlay.sourceNote.localId]
        val targetPos: Point = when {
            target != null -> target.noteView.getLocationOnScreenAsPoint()
            else -> {
                val index = to.notesList.notesCollection.indexOfFirst { it.localId == note.localId }
                overlay.getLocationOnScreenAsPoint() + when {
                    index < firstTargetViewIndex -> Point(0, -overlayHeight)
                    else -> Point(0, overlayHeight)
                }
            }
        } - overlayPosition

        return animatePosition(overlay, targetPos)
    }

    private fun createResultOverlays(from: SearchFragment, activity: FragmentActivity,
                                     overlayPosition: Point, animationOverlay: ViewGroup): List<NoteItemComponent> {
        val fromRecyclerView = from.notesList!!.notesRecyclerView

        val noteOverlays = fromRecyclerView.childrenSequence().toList().map {
            val srcView = it as NoteItemComponent
            createNoteItemOverlay(activity, srcView.sourceNote, animationOverlay).apply {
                setLayoutParamsFrom(srcView, overlayPosition)
            }
        }
        return noteOverlays
    }

    private fun createNoteItemOverlay(activity: FragmentActivity, note: Note, animationOverlay: ViewGroup)
            : NoteItemComponent {
        val layoutId = when {
            note.isMediaListEmpty -> R.layout.note_item_layout_text
            note.hasNoText -> R.layout.note_item_layout_image
            else -> R.layout.note_item_layout_mixed
        }
        val view = LayoutInflater.from(activity).inflate(layoutId, animationOverlay, false)
                as NoteItemComponent
        view.bindNote(note, false)
        animationOverlay.addView(view)
        return view
    }

    private fun animateNotesListFadeAfterResultsMove(view: View) =
            fadeIn(view).apply {
                duration = RESULTS_TO_NOTES_LIST_FADE_DURATION
                startDelay = RESULTS_TO_NOTES_LIST_FADE_DELAY
                decelerate()
            }

    private fun animatePosition(view: View, endPos: Point) =
            ObjectAnimator.ofPropertyValuesHolder(view,
                    PropertyValuesHolder.ofFloat(View.X, view.x, endPos.x.toFloat()),
                    PropertyValuesHolder.ofFloat(View.Y, view.y, endPos.y.toFloat()))

    private fun fadeIn(view: View) =
            ObjectAnimator.ofFloat(view, View.ALPHA, view.alpha, 1f)

    private fun slideAndFadeIn(context: Context) =
            ScaleTranslateAndFadeTransition(
                    dy = context.resources.getDimension(R.dimen.color_filter_translate_animation)).decelerate()

    private fun shrinkAndSlideOut(context: Context) =
            ScaleTranslateAndFadeTransition(
                    dy = context.resources.getDimension(R.dimen.color_filter_translate_animation), scale = .85f)
                    .accelerate()

    private fun clipOverlayToView(activity: FragmentActivity, view: View, overlayPosition: Point,
                                  animationOverlay: ViewGroup) {
        animationOverlay.clipBounds = view.getBoundsOnScreen() - overlayPosition
    }
}